program HopNet;

{ A 7-node, symmetrically weighted Hopfield network.
  Based upon network in Communications of the ACM article,
  November 1990, "Connectionist Ideas and Algorithms".

  Author: David J Benn
    Date: 8th-15th March 1991 }

const xoff=20;
      yoff=35;
      radius=10;
      on=1;
      off=0;

type  strng = string[3];

var   coords,
      weights : text;

      nodex,
      nodey      : array [0..6] of integer;
      node       : array [0..6] of boolean;
      wt,
      initial_wt : array [0..6,0..6] of integer;

      outline,
      quit       : boolean;

{------------------------------------------------}

procedure intro;
begin
     hires;
     writeln('       *** HopNet v1.2 ***');
end;

procedure read_coords;
var i,x,y : integer;
begin
     {$I-}
     assign(coords,'coords.dat');
     reset(coords);
     for i:=0 to 6 do
     begin
          readln(coords,x,y);
          nodex[i]:=x; nodey[i]:=y;
          nodex[i]:=nodex[i] + xoff;
          nodey[i]:=nodey[i] + yoff;
     end;
     close(coords);
     if (IOresult<>0) then
     begin
          writeln('COORDS.DAT not found!');
          halt;
     end;
     {$I+}
end;

procedure read_weights;
var i,j,w : integer;
begin
     {$I-}
     assign(weights,'weights.dat');
     reset(weights);
     for i:=0 to 6 do
     begin
          for j:=0 to 6 do
          begin
               read(weights,w);
               wt[i,j]:=w;
               initial_wt[i,j]:=wt[i,j];
          end;
          readln(weights);
     end;
     close(weights);
     if IOresult<>0 then
     begin
          writeln('WEIGHTS.DAT not found!');
          halt;
     end;
     {$I+}
end;

procedure reset_nodes;
var i : integer;
begin
     for i:=0 to 6 do
         node[i]:=false;
end;

procedure square(xc,yc : integer);
var x,y : integer;
begin
     draw(xc-radius*2,yc-radius,xc-radius*2,yc+radius,1);
     draw(xc+radius*2,yc-radius,xc+radius*2,yc+radius,1);
     draw(xc-radius*2,yc-radius,xc+radius*2,yc-radius,1);
     draw(xc-radius*2,yc+radius,xc+radius*2,yc+radius,1);
end;

procedure fill_square(xc,yc,colr : integer);
var x1,x2,y : integer;
begin
     for y:=yc-radius + 1 to yc+radius - 1 do
     begin
          x1:=xc-radius*2 + 1;
          x2:=xc+radius*2 - 1;
          Draw(x1,y,x2,y,colr);
     end;
end;

procedure clrline;
begin
     gotoxy(52,15);
     write('                             ');
end;

procedure clrchoice;
begin
     gotoxy(64,13);
     write('         ');
end;

procedure plot_link(n1,n2 : integer; connected : boolean);
var colr : integer;
begin
     if connected then colr:=1 else colr:=0;
     draw(nodex[n1],nodey[n1],nodex[n2],nodey[n2],colr);
     if node[n1] then
        fill_square(nodex[n1],nodey[n1],on)
     else fill_square(nodex[n2],nodey[n2],off);
     if node[n2] then
        fill_square(nodex[n2],nodey[n2],on)
     else fill_square(nodex[n2],nodey[n2],off);
end;

procedure plot_node(n : integer; active : boolean);
begin
     if outline then square(nodex[n],nodey[n]);
     if active then
        fill_square(nodex[n],nodey[n],on)
     else
        fill_square(nodex[n],nodey[n],off);
end;

procedure plot_weight(i,j : integer);
var x,y,k  : integer;
    weight : strng;
begin
     x:=(nodex[i]+nodex[j]) div 16 + 1;
     y:=(nodey[i]+nodey[j]) div 16 + 1;
     gotoxy(x,y);
     str(wt[i,j],weight);
     if weight[1]<>'-' then insert('+',weight,1);
     for k:=1 to length(weight) do
         if weight[k]=' ' then
            write('+')
         else
            write(weight[k]);
end;

procedure draw_links;
var i,j,bse : integer;
begin
     bse:=0;
     for i:=0 to 6 do
     begin
          for j:=bse to 6 do
              if wt[i,j]<>0 then draw(nodex[i],nodey[i],nodex[j],nodey[j],1);
          bse:=bse+1;
     end;
end;

procedure draw_nodes;
var i : integer;
begin
     outline:=true;
     for i:=0 to 6 do
         plot_node(i,node[i]);
     outline:=false;
end;

procedure draw_weights;
var i,j,bse : integer;
begin
     bse:=0;
     for i:=0 to 6 do
     begin
          for j:=bse to 6 do
              if wt[i,j]<>0 then plot_weight(i,j);
          bse:=bse+1;
     end;
end;

procedure node_key;
begin
     gotoxy(38,5);  write('0');
     gotoxy(42,5);  write('1');
     gotoxy(36,7);  write('2');
     gotoxy(44,7);  write('3');
     gotoxy(37,9);  write('4');
     gotoxy(40,10); write('5');
     gotoxy(43,9);  write('6');
end;

procedure blank_node_key;
begin
     gotoxy(38,5);  write(' ');
     gotoxy(42,5);  write(' ');
     gotoxy(36,7);  write(' ');
     gotoxy(44,7);  write(' ');
     gotoxy(37,9);  write(' ');
     gotoxy(40,10); write(' ');
     gotoxy(43,9);  write(' ');
end;

procedure show_matrix;
var i,j,bse : integer;
begin
     for i:=0 to 6 do
     begin
          gotoxy(52+i*4,17);
          write(i:4);
     end;
     for j:=0 to 6 do
     begin
          gotoxy(51,19+j);
          write(j);
     end;
     {values}
     bse:=0;
     for i:=0 to 6 do
     begin
          for j:=bse to 6 do
          begin
               gotoxy(52+4*j,19+i);
               write(wt[i,j]:4);
          end;
          bse:=bse+1;
     end;
end;

procedure set_node;
var n  : integer;
begin
     {$I-}
     repeat
           clrline;
           gotoxy(52,15);
           write('set which node? (0..6) ');
           readln(n);
     until(n>=0) and (n<=6) and (IOresult=0);
     {$I+}
     node[n]:=true;
     fill_square(nodex[n],nodey[n],on);
     {in case any weights obscured...}
     draw_weights;
end;

procedure reset_node;
var n  : integer;
begin
     {$I-}
     repeat
           clrline;
           gotoxy(52,15);
           write('reset which node? (0..6) ');
           readln(n);
     until(n>=0) and (n<=6) and (IOresult=0);
     {$I+}
     node[n]:=false;
     fill_square(nodex[n],nodey[n],off);
     draw_weights;
end;

procedure change_weight;
var n1,n2,x,y,weight : integer;
begin
     {$I-}
   repeat
     repeat
           clrline;
           gotoxy(52,15);
           write('first node? (0..6) ');
           readln(n1);
     until (n1>=0) and (n1<=6) and (IOresult=0);
     repeat
           clrline;
           gotoxy(52,15);
           write('second node? (0..6) ');
           readln(n2);
     until (n2>=0) and (n2<=6) and (IOresult=0);
   until (n1<>n2);
     repeat
           clrline;
           gotoxy(52,15);
           write('connection weight? ');
           readln(weight);
     until (IOresult=0);
     {$I+}
     if weight<>0 then
     begin
          wt[n1,n2]:=weight;
          wt[n2,n1]:=weight;
          intro;
          draw_links;
          draw_nodes;
          draw_weights;
          node_key;
          show_matrix;
     end
     else
     begin
          wt[n1,n2]:=0;
          wt[n2,n1]:=0;
          intro;
          draw_links;
          draw_nodes;
          draw_weights;
          node_key;
          show_matrix;
     end;
end;

procedure reset_all_nodes;
var i : integer;
begin
     gotoxy(52,15);
     write('resetting nodes...');
     for i:=0 to 6 do
     begin
          node[i]:=false;
          fill_square(nodex[i],nodey[i],off);
     end;
     draw_weights;
end;

procedure reset_all_weights;
var i,j : integer;
begin
     for i:=0 to 6 do
     begin
          for j:=0 to 6 do
          wt[i,j]:=initial_wt[i,j];
     end;
     intro;
     draw_links;
     draw_nodes;
     draw_weights;
     node_key;
     show_matrix;
end;

procedure rnd_nodes;
var i,r : integer;
begin
     gotoxy(52,15);
     write('randomizing nodes...');
     for i:=0 to 6 do
     begin
          r:=random(2);
          if r=0 then
          begin
               node[i]:=true;
               plot_node(i,true);
          end
          else
              begin
                   node[i]:=false;
                   plot_node(i,false);
              end;
     end;
     draw_weights;
end;

procedure show_node_data;
var i : integer;
begin
     gotoxy(32,3); write('node:');
     gotoxy(42,3); write('sum:');
     for i:=0 to 6 do
     begin
          gotoxy(32,4+i);
          write(i:4);
     end;
end;

procedure show_node_sum(node,sum : integer);
begin
     gotoxy(42,4+node); write(sum:3);
end;

procedure clr_node_data;
var i : integer;
begin
     for i:=0 to 8 do
     begin
          gotoxy(32,3+i);
          write('                ');
     end;
end;

procedure seek;
var pass,k,i,j,r : integer;
    sum,
    oldsum : array [0..6] of integer;
    flag   : array [0..6] of boolean;
    step   : char;
    change : boolean;
begin
     {clear sums}
     for i:=0 to 6 do begin sum[i]:=0; oldsum[i]:=0; end;
        repeat
          clrline;
          gotoxy(52,15);
          write('stop after each pass? (y/n)');
          readln(step);
          step:=Upcase(step);
        until (step='Y') or (step='N');
     clrline;
     blank_node_key;
     show_node_data;
     pass:=1;
     repeat
          gotoxy(52,15);
          write('pass #',pass);
          for k:=0 to 6 do flag[k]:=false;
          {do one pass}
          change:=false;
          for i:=0 to 6 do
          begin
               repeat r:=random(7); until not(flag[r]);   {0..6}
               sum[r]:=0;
               flag[r]:=true;
               {linked to any other nodes? are they active?}
               for j:=0 to 6 do
                    if wt[r,j]<>0 then
                       if node[j] then sum[r]:=sum[r]+wt[r,j];
               show_node_sum(r,sum[r]);
               if sum[r]>0 then
               begin
                    node[r]:=true;
                    plot_node(r,true);
               end
               else
                   begin
                        node[r]:=false;
                        plot_node(r,false);
                   end;
               if KeyPressed then
               begin
                    clrline;
                    clr_node_data;
                    node_key;
                    exit;
               end;
               {has this node changed?}
               if not(change) then
                    if sum[r]<>oldsum[r] then change:=true;
               oldsum[r]:=sum[r];
               draw_weights;
          end;
          if (step='Y') and (change) then
          begin
               gotoxy(52,15);
               write('press any key...');
               while not(KeyPressed) do;
               clrline;
         end;
         pass:=pass+1;
    until not(change);
     gotoxy(52,15);
     write('network stable. hit a key...');
     while not(KeyPressed) do;
     clrline;
     clr_node_data;
     node_key;
end;

procedure zero_all_weights;
var i,j : integer;
begin
     for i:=0 to 6 do
     begin
          for j:=0 to 6 do
          wt[i,j]:=0;
     end;
     intro;
     draw_links;
     draw_nodes;
     draw_weights;
     node_key;
     show_matrix;
end;

procedure set_pattern;
{incorporate pattern into network}
var i,j : integer;
begin
     for i:=0 to 6 do
     if node[i] then
     begin
        for j:=0 to 6 do
          if i<>j then
           {if node active -> excitatory or inhibitory link to others}
           if node[j] then wt[i,j]:=wt[i,j]+1
           else
               if not(node[j]) then wt[i,j]:=wt[i,j]-1;
     end;
     {make sure matrix is symmetrical}
     for i:=0 to 6 do
         for j:=0 to 6 do
              if i<>j then
                 if wt[i,j]<>0 then
                    wt[j,i]:=wt[i,j]
                 else
                 if wt[j,i]<>0 then
                    wt[i,j]:=wt[j,i];
     intro;
     draw_links;
     draw_nodes;
     draw_weights;
     node_key;
     show_matrix;
end;

procedure help;
begin
  clrscr;
  writeln('HopNet v1.2 -- David Benn  1991 (University of Tasmania - Launceston)');
  writeln;
  writeln(' This program is based on the 7-node Hopfield neural network in');
  writeln('the article: "Connectionist Ideas and Algorithms", Communications');
  writeln('of the ACM, November 1990. In addition, the article: "Collective');
  writeln('Computation in Neuronlike Circuits", Scientific American, December');
  writeln('1987 provided further information re: associative memory.');
  writeln;
  writeln(' Each square represents a node and each line a symmetrical');
  writeln('connection between two nodes. The numbers on each line indicate');
  writeln('the weight (strength) of the connection.');
  writeln;
  writeln(' A node is activated if it is filled. Changing a weight to zero');
  writeln('causes a physical (and logical) link to be removed, while a new');
  writeln('connection can be established by specifying a non-zero weight for');
  writeln('two unconnected nodes. Also, an existing weight can be modified.');
  writeln;
  writeln(' In seeking a stable state, HopNet indicates each pass (one');
  writeln('complete check of all 7 nodes) and the sums for each node.');
  writeln('The process stops when no summations are able to change.');
  writeln('A sequence of passes can be cancelled by hitting any key.');
  writeln;
  writeln('press any key...');
  while not(KeyPressed) do;
  clrscr;
  writeln;
  writeln(' Option 8 allows for the arbitrary setting of nodes and their');
  writeln('incorporation into the network as a pattern, by automatic setting');
  writeln('of the appropriate weights. It is best to zero all weights before');
  writeln('using this option.');
  writeln;
  writeln(' Hopnet - consisting of HOPNET.PAS, HOPNET.COM, COORDS.DAT');
  writeln('and WEIGHTS.DAT - is freely distributable and was written');
  writeln('in Turbo Pascal version 3.');
  writeln;
  writeln(' Communicate suggestions, gripes or bugs to me on 317 680 (home).');
  writeln('                                                [15th March 1991]');
  writeln;
  writeln('press any key for menu...');
  while not(KeyPressed) do;
  intro;
  draw_links;
  draw_nodes;
  draw_weights;
  node_key;
  show_matrix;
end;

procedure menu;
var choice : integer;
begin
     {$I-}
     repeat
           clrchoice;
           gotoxy(52,1);  write(' 1. activate a node');
           gotoxy(52,2);  write(' 2. deactivate a node');
           gotoxy(52,3);  write(' 3. randomize nodes');
           gotoxy(52,4);  write(' 4. reset all nodes');
           gotoxy(52,5);  write(' 5. change a weight');
           gotoxy(52,6);  write(' 6. zero all weights');
           gotoxy(52,7);  write(' 7. reset all weights');
           gotoxy(52,8);  write(' 8. incorporate a pattern');
           gotoxy(52,9);  write(' 9. seek a stable state');
           gotoxy(52,10); write('10. help');
           gotoxy(52,11); write(' 0. quit');
           gotoxy(52,13); write('enter 0..10: ');
           read(choice);
     until (choice>=0) and (choice<=10) and (IOresult=0);
     {$I+}
     case choice of
          0 : begin quit:=true; clrscr; end;
          1 : set_node;
          2 : reset_node;
          3 : rnd_nodes;
          4 : reset_all_nodes;
          5 : change_weight;
          6 : zero_all_weights;
          7 : reset_all_weights;
          8 : set_pattern;
          9 : seek;
         10 : help;
     end;
     clrline;
     clrchoice;
end;

{Main}
begin
     read_coords;
     read_weights;
     intro;
     reset_nodes;
     draw_links;
     draw_nodes;
     draw_weights;
     node_key;
     show_matrix;
     quit:=false;
     while not(quit) do menu;
end.



